"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
require("bluebird-global");
const sqlite_1 = __importDefault(require("sqlite"));
const emittery_1 = __importDefault(require("emittery"));
const sql_template_strings_1 = require("sql-template-strings");
const table_1 = require("./table");
const find_1 = require("./find");
const utils_1 = require("./utils");
const sql_template_strings_2 = require("./compat/sql-template-strings");
class Db extends emittery_1.default.Typed {
    constructor(params) {
        super();
        this.sql = params.sql;
    }
    static async connect(f, options) {
        const sql = typeof f === 'string' ? await sqlite_1.default.open(f, options) : f;
        return new Db({ sql });
    }
    /**
     * Initialize tables sequentially, just in case foreign keys matter
     *
     * @param tables
     */
    async init(tables) {
        for (const t of tables) {
            await t.__init(this.sql);
        }
    }
    create(table) {
        return table.create(this.sql);
    }
    each(table0, ...tables) {
        return (qCond, select, options = {}) => {
            return async (cb) => {
                return (await this._find(table0, ...tables)(qCond, select, options)).each(cb);
            };
        };
    }
    all(table0, ...tables) {
        return async (qCond, select, options = {}) => {
            return (await this._find(table0, ...tables)(qCond, select, options)).all();
        };
    }
    first(table0, ...tables) {
        return async (qCond, select, options = {}) => {
            return (await this._find(table0, ...tables)(qCond, select, options)).first();
        };
    }
    update(table0, ...tables) {
        return async (qCond, set, options = {}) => {
            await Promise.all((await this._findIds(table0, ...tables)(qCond, options)).map(async ({ sql, table }) => {
                if (Array.isArray(set)) {
                    await Promise.all(set.filter((s) => s.table === table).map(async (s) => {
                        await table.__updateBySql(this.sql)(sql, s.set);
                    }));
                }
                else {
                    await table.__updateBySql(this.sql)(sql, set);
                }
            }));
        };
    }
    delete(table0, ...tables) {
        return async (qCond, options = {}) => {
            await Promise.all((await this._findIds(table0, ...tables)(qCond, options)).map(async ({ sql, table }) => {
                await table.__deleteBySql(this.sql)(sql);
            }));
        };
    }
    async close() {
        await this.sql.close();
        return this;
    }
    _find(table0, ...tables) {
        return async (qCond, select, options = {}) => {
            const tablesArray = tables.map((t) => t instanceof table_1.Table ? { to: t } : t);
            await this.emit('pre-find', {
                cond: qCond,
                select: select === '*' ? table0.c : select,
                tables: tablesArray,
                options,
            });
            const selectDict = Object.entries(select === '*' ? table0.c : select).map(([alias, col]) => {
                const key = col instanceof table_1.Column
                    ? utils_1.safeColumnName(`${col.tableName}.${col.name}`)
                    : col instanceof sql_template_strings_1.SQLStatement
                        ? col.text : utils_1.safeColumnName(col);
                return [alias, {
                        key,
                        column: col instanceof table_1.Column ? col : undefined,
                    }];
            }).reduce((prev, [a, k]) => ({ ...prev, [a]: k }), {});
            const tableRecord = [
                table0,
                ...tables.map((t) => t instanceof table_1.Table
                    ? t
                    : t.to instanceof table_1.Table ? t.to : t.to.opts.table),
            ].reduce((prev, t) => ({ ...prev, [t.m.__meta.name]: t }), {});
            const postfix = options.postfix ? [sql_template_strings_2.SQL(options.postfix)] : [];
            if (options.sort) {
                postfix.push(sql_template_strings_2.SQL `ORDER BY ${sql_template_strings_2.SQL(utils_1.safeColumnName(typeof options.sort.key === 'string' ? options.sort.key : `${options.sort.key.tableName}.${options.sort.key.name}`))} ${sql_template_strings_2.SQL(options.sort.desc ? 'DESC' : 'ASC')}`);
            }
            if (options.limit) {
                postfix.push(sql_template_strings_2.SQL(`LIMIT ${options.limit}`));
            }
            if (options.offset) {
                postfix.push(sql_template_strings_2.SQL(`OFFSET ${options.offset}`));
            }
            const sql = sql_template_strings_2.joinSQL([
                `SELECT ${selectDict ? Object.entries(selectDict).map(([a, k]) => {
                    return k.key === a ? k.key : `${k.key} AS ${utils_1.safeColumnName(a)}`;
                }).join(',') : select}`,
                `FROM ${utils_1.safeColumnName(table0.m.__meta.name)}`,
                ...tablesArray.map((t) => {
                    const toTable = t.to instanceof table_1.Table ? t.to : t.to.opts.table;
                    if (t.from) {
                        return [
                            `${t.type || 'INNER'} JOIN ${utils_1.safeColumnName(toTable.m.__meta.name)}`,
                            'ON',
                            utils_1.safeColumnName(`${t.from.tableName}.${t.from.name}`),
                            '=',
                            utils_1.safeColumnName(t.to instanceof table_1.Table
                                ? `${toTable.m.__meta.name}.${toTable.primaryKey}`
                                : `${toTable.m.__meta.name}.${t.to.name}`),
                        ].join(' ');
                    }
                    else if (t.cond) {
                        return [
                            `${t.type || 'INNER'} JOIN ${utils_1.safeColumnName(toTable.m.__meta.name)}`,
                            'ON',
                            `(${t.cond})`,
                        ].join(' ');
                    }
                    else {
                        return `${t.type || 'NATURAL'} JOIN ${utils_1.safeColumnName(toTable.m.__meta.name)}`;
                    }
                }),
                sql_template_strings_2.SQL `WHERE ${find_1.parseCond(qCond, tableRecord)}`,
                ...postfix,
            ], ' ');
            await this.emit('find-sql', sql);
            const parseNative = (r) => {
                Object.entries(r).map(([alias, v]) => {
                    if (selectDict && selectDict[alias] && selectDict[alias].column) {
                        const col = selectDict[alias].column;
                        r[alias] = col.opts.table.transform(col.name, 'get')(v);
                    }
                });
                return r;
            };
            return {
                sql,
                first: async () => {
                    const r = await this.sql.get(sql);
                    return r ? parseNative(r) : null;
                },
                each: (cb) => {
                    return new Promise((resolve, reject) => {
                        this.sql.each(sql, (err, r) => {
                            if (err) {
                                reject(err);
                            }
                            // eslint-disable-next-line standard/no-callback-literal
                            cb(parseNative(r));
                        }).then((n) => resolve(n));
                    });
                },
                all: async () => (await this.sql.all(sql)).map((r) => parseNative(r)),
            };
        };
    }
    _findIds(table0, ...tables) {
        return async (qCond, options = {}) => {
            const tt = [
                table0,
                ...tables.map((t) => t instanceof table_1.Table
                    ? t
                    : t.to instanceof table_1.Table
                        ? t.to : t.to.opts.table),
            ];
            return Promise.all(tt.map(async (t) => {
                return {
                    sql: (await this._find(table0, ...tables)(qCond, {
                        [`${t.m.__meta.name}__id`]: new table_1.Column({
                            name: t.primaryKey,
                            table: t,
                            prop: t.m.__meta.primary,
                        }),
                    }, options)).sql,
                    table: t,
                };
            }));
        };
    }
}
exports.Db = Db;
