"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const emittery_1 = __importDefault(require("emittery"));
const sql_template_strings_1 = require("sql-template-strings");
const utils_1 = require("./utils");
const sql_template_strings_2 = require("./compat/sql-template-strings");
class Column {
    constructor(opts) {
        this.opts = opts;
    }
    get tableName() {
        return this.opts.table.m.__meta.name;
    }
    get name() {
        return this.opts.name;
    }
}
exports.Column = Column;
class Table extends emittery_1.default.Typed {
    constructor(M) {
        super();
        this.m = new M();
        if (this.m.__meta.createdAt) {
            this.m.__meta.prop.createdAt = { type: 'Date', null: false, default: () => new Date() };
        }
        if (this.m.__meta.updatedAt) {
            this.m.__meta.prop.updatedAt = {
                type: 'Date',
                null: false,
                onChange: () => new Date(),
            };
        }
        this.c = Object.entries(this.m.__meta.prop).map(([k, v]) => {
            if (v) {
                return [k, new Column({
                        name: k,
                        table: this,
                        prop: v,
                    })];
            }
            return null;
        }).filter((el) => el)
            .reduce((prev, [k, v]) => ({ ...prev, [k]: v }), {});
        this.c[this.primaryKey] = new Column({
            name: this.primaryKey,
            table: this,
            prop: this.m.__meta.primary,
        });
        Object.entries(this.m.__meta.prop).map(([k, v]) => {
            if (v) {
                const { default: def, onChange, onUpdate } = v;
                if (typeof def === 'function' || onChange !== undefined) {
                    const fn = def || onChange;
                    this.on('pre-create', async ({ entry }) => {
                        if (utils_1.isNullOrUndefined(entry[k])) {
                            entry[k] = typeof fn === 'function' ? await fn(entry) : fn;
                        }
                    });
                }
                if (onUpdate !== undefined || onChange !== undefined) {
                    const fn = onUpdate !== undefined ? onUpdate : onChange;
                    this.on('pre-update', async ({ set }) => {
                        /**
                         * NULL should be able to set SQLite row to BLANK
                         */
                        if (set[k] === undefined) {
                            set[k] = typeof fn === 'function' ? await fn(set) : fn;
                        }
                    });
                }
            }
        });
    }
    get primaryKey() {
        return this.m.__meta.primary && typeof this.m.__meta.primary.name === 'string' ? this.m.__meta.primary.name : 'ROWID';
    }
    get name() {
        return this.m.__meta.name;
    }
    async __init(db) {
        const getDefault = (k, v) => {
            if (utils_1.isNullOrUndefined(v.default)) {
                return sql_template_strings_2.SQL();
            }
            else if (v.default instanceof sql_template_strings_1.SQLStatement) {
                return sql_template_strings_2.SQL `DEFAULT ${v.default}`;
            }
            else if (typeof v.default === 'string') {
                return sql_template_strings_2.SQL(`DEFAULT '${v.default.replace(/'/, "[']")}'`);
            }
            else if (typeof v.default === 'number') {
                return sql_template_strings_2.SQL(`DEFAULT ${v.default}`);
            }
            else if (typeof v.default === 'boolean') {
                return sql_template_strings_2.SQL(`DEFAULT ${v.default ? 1 : 0}`);
            }
            return sql_template_strings_2.SQL();
        };
        const cols = [];
        if (this.m.__meta.primary && this.m.__meta.primary.type) {
            cols.push(sql_template_strings_2.joinSQL([
                utils_1.safeColumnName(this.m.__meta.primary.name),
                utils_1.AliasToSqliteType[this.m.__meta.primary.type] || 'INTEGER',
                'PRIMARY KEY',
                ...(this.m.__meta.primary.autoincrement ? [
                    'AUTOINCREMENT',
                ] : []),
                getDefault(this.m.__meta.primary.name, this.m.__meta.primary),
            ], ' '));
        }
        for (const [k, v] of Object.entries(this.m.__meta.prop)) {
            if (v && v.type) {
                cols.push(sql_template_strings_2.joinSQL([
                    utils_1.safeColumnName(k),
                    utils_1.AliasToSqliteType[v.type] || 'TEXT',
                    ...(v.null ? [] : [
                        'NOT NULL',
                    ]),
                    getDefault(k, v),
                    ...(v.references ? [
                        `REFERENCES ${utils_1.safeColumnName(v.references)}`,
                    ] : []),
                ], ' '));
            }
        }
        if (this.m.__meta.primary && Array.isArray(this.m.__meta.primary.name)) {
            cols.push(sql_template_strings_2.SQL `PRIMARY KEY (${sql_template_strings_2.joinSQL(this.m.__meta.primary.name.map((k) => utils_1.safeColumnName(k)), ',')})`);
        }
        if (this.m.__meta.unique && this.m.__meta.unique.length > 0) {
            this.m.__meta.unique.forEach((ss) => {
                cols.push(sql_template_strings_2.SQL `CONSTRAINT ${sql_template_strings_2.SQL(ss.name)} UNIQUE (${sql_template_strings_2.joinSQL(ss.keys.map((k) => utils_1.safeColumnName(k)), ',')})`);
            });
        }
        for (const [k, v] of Object.entries(this.m.__meta.prop)) {
            if (v && v.unique) {
                cols.push(sql_template_strings_2.SQL `CONSTRAINT ${sql_template_strings_2.SQL(utils_1.safeColumnName(v.unique))} UNIQUE (${sql_template_strings_2.SQL(utils_1.safeColumnName(k))})`);
            }
        }
        const sql = sql_template_strings_2.SQL `CREATE TABLE IF NOT EXISTS ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.m.__meta.name))} (${sql_template_strings_2.joinSQL(cols, ',')})`;
        await this.emit('build-sql', sql);
        await db.run(sql);
        if (this.m.__meta.index) {
            await Promise.all(this.m.__meta.index.map(async (idx) => {
                const sql = sql_template_strings_2.SQL `CREATE INDEX IF NOT EXISTS ${sql_template_strings_2.SQL(utils_1.safeColumnName(idx.name))} ON ${this.m.__meta.name} (${sql_template_strings_2.joinSQL(idx.keys.map((k) => sql_template_strings_2.SQL(utils_1.safeColumnName(k))), ',')})`;
                await this.emit('build-sql', sql);
                await db.run(sql);
            }));
        }
        for (const [k, v] of Object.entries(this.m.__meta.prop)) {
            if (v && v.index) {
                const sql = sql_template_strings_2.SQL `CREATE INDEX IF NOT EXISTS ${sql_template_strings_2.SQL(utils_1.safeColumnName(v.index))} ON ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.m.__meta.name))} (${sql_template_strings_2.SQL(utils_1.safeColumnName(k))})`;
                await this.emit('build-sql', sql);
                await db.run(sql);
            }
        }
    }
    create(db) {
        return async (entry, options = {}) => {
            const postfix = options.postfix ? [options.postfix] : [];
            if (options.ignoreErrors) {
                postfix.push(sql_template_strings_2.SQL `ON CONFLICT DO NOTHING`);
            }
            await this.emit('pre-create', { entry, options: { postfix } });
            const keys = [];
            const values = [];
            for (const [k, v] of Object.entries(entry)) {
                if (typeof v !== 'undefined') {
                    keys.push(k);
                    values.push(this.transform(k, 'set')(v));
                }
            }
            const sql = sql_template_strings_2.SQL `INSERT INTO ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.m.__meta.name))} (${sql_template_strings_2.joinSQL(keys.map((k) => utils_1.safeColumnName(k)), ',')}) VALUES (${sql_template_strings_2.joinSQL(values.map((v) => sql_template_strings_2.SQL `${v}`), ',')}) ${sql_template_strings_2.joinSQL(postfix, ' ')}`;
            await this.emit('create-sql', sql);
            const r = await db.run(sql);
            return r.lastID;
        };
    }
    __updateBySql(db) {
        return async (sql, set) => {
            await this.emit('pre-update', { sql, set });
            const setSql = [];
            for (const [k, v] of Object.entries(set)) {
                if (typeof v !== 'undefined') {
                    setSql.push(sql_template_strings_2.SQL `${sql_template_strings_2.SQL(utils_1.safeColumnName(k))} = ${this.transform(k, 'set')(v)}`);
                }
            }
            const resultSql = sql_template_strings_2.SQL `UPDATE ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.name))} SET ${sql_template_strings_2.joinSQL(setSql, ',')} WHERE ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.primaryKey))} IN (${sql})`;
            await this.emit('update-sql', resultSql);
            await db.run(resultSql);
        };
    }
    __deleteBySql(db) {
        return async (sql) => {
            await this.emit('pre-delete', { sql });
            const resultSql = sql_template_strings_2.SQL `DELETE FROM ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.name))} WHERE ${sql_template_strings_2.SQL(utils_1.safeColumnName(this.primaryKey))} IN (${sql})`;
            await this.emit('delete-sql', resultSql);
            await db.run(resultSql);
        };
    }
    /**
     * @internal
     * @param k
     * @param method
     */
    transform(k, method = 'set') {
        let fn = null;
        const prop = this.m.__meta.prop[k];
        if (prop) {
            if (prop.transform) {
                fn = prop.transform[method] || null;
            }
            if (!fn) {
                const t = exports._transformers[prop.type];
                if (t) {
                    fn = t[method] || null;
                }
            }
        }
        return fn || ((a) => a);
    }
}
exports.Table = Table;
exports._transformers = {
    Date: {
        get: (repr) => typeof repr === 'number' ? new Date(repr) : null,
        set: (d) => d ? d instanceof Date ? +d : +new Date(d) : null,
    },
    JSON: {
        get: (repr) => repr ? JSON.parse(repr) : null,
        set: (data) => data ? JSON.stringify(data) : null,
    },
    StringArray: {
        get: (repr) => (() => {
            repr = repr ? repr.substr(1, repr.length - 2) : '';
            return repr ? repr.split('\x1f') : null;
        })(),
        set: (d) => d ? '\x1f' + d.join('\x1f') + '\x1f' : null,
    },
    Boolean: {
        get: (repr) => typeof repr === 'number' ? repr !== 0 : null,
        set: (d) => typeof d === 'boolean' ? Number(d) : null,
    },
};
